#include <iostream>
#include <fstream>
#include <set>
#include <cmath>
#include <boost/shared_ptr.hpp>
#include "include/linalg.hpp"
#include "include/rbf.hpp"
#include "include/bvp.hpp"
#include "include/error.hpp"
#include "include/utils.hpp"
#include "include/diff_op.hpp"
#include "include/ddm.hpp"

using namespace kwantxi;

class conv_diff : public linear_diff_op2{
public:
  conv_diff(double u_in) : u(u_in) {} ;
  double at(double (*f)(const point &p), const point &p) const;
  double at(const radial_basis_function &RBF, const point &p) const;

  void set_u(double u_in){
    u = u_in;
  };

private:
  Laplacian L;
  del1 D;
  double u;
};

double conv_diff::at(double (*f)(const point &p), const point &p) const{
  return L.at(f,p) - u*D.at(f,p);
}

double conv_diff::at(const radial_basis_function &RBF, const point &p) const{
  return L.at(RBF,p) - u*D.at(RBF,p);
}


//New diff_op requested by Moreles...
class smoothed_conv_diff : public linear_diff_op2{
public:

  smoothed_conv_diff(double u_in, double h_in, const vector& n_in) : 
    h(h_in), n(n_in) {
    L1.set_u(u_in);
  };

  double at(double (*f)(const point &p), const point &p) const;
  double at(const radial_basis_function &RBF, const point &p) const;

private:
  point gradL(const radial_basis_function& RBF, const point& p)const;
  static const radial_basis_function* the_rbf;
  static double wrapper(double x, void* nothing);
  static size_t direction;
  static point x;
  static conv_diff L1;
  double h;
  vector n;
};

//Declare statics...
const radial_basis_function* smoothed_conv_diff::the_rbf = 0;
point smoothed_conv_diff::x(2);
conv_diff smoothed_conv_diff::L1(1);
size_t smoothed_conv_diff::direction = 1;


double smoothed_conv_diff::
at(const radial_basis_function &RBF, const point &p) const{
  the_rbf = &RBF;
  return L1.at(RBF,p) + h*(n*gradL(RBF,p));
}

double smoothed_conv_diff::
at(double (*f)(const point &p), const point &p) const{
  f = 0;
  p(1);
  return 0;
}

point smoothed_conv_diff::
gradL(const radial_basis_function& RBF, const point& p)const{

  point result(2);
  the_rbf = &RBF;
  sqrteps = sqrt(std::numeric_limits<double>::epsilon());

  for(size_t i = 1; i <=2; i++){
    direction = i;
    double diff_i,  abserror;
    double t = p(i); 
    double typt = (1 > log(t) ? 1 : log(t));
    double step = sqrteps*( fabs(t) > typt ? fabs(t) : typt);
    gsl_function *f = new gsl_function;
    f -> function = &wrapper;
    f -> params = 0;
    gsl_deriv_central(f, t, step, &diff_i, &abserror);
    delete f;
    result(i) = diff_i;
  }
  return result;
}

double smoothed_conv_diff::
wrapper(double xi, void* nothing){
  nothing = 0;
  x(direction) = xi;
  return L1.at(*the_rbf, x);
}

//Now for the boundary operators...

class bdry_conv_diff : public bdry_diff_op{
public:
  double at(double (*f)(const point &p), const point &p) const{
    if(p(1) == 0 or p(1) == 1)
      return D(f,p);
    return N(f,p);
  };
  double at(double (*f)(const point &p), const point &p, 
	    const vector &n) const{
    if(p(1) == 0 or p(1) == 1)
      return D(f,p,n);
    return N(f,p,n);
  };
  double at(const radial_basis_function &RBF, const point &p) const{
    if(p(1) == 0 or p(1) == 1)
      return D(RBF,p);
    return N(RBF,p);
  };
  double at(const radial_basis_function &RBF, const point &p, 
	    const vector &n) const{
    if(p(1) == 0 or p(1) == 1)
      return D(RBF,p,n);
    return N(RBF,p,n);
  };
private:
  neumann_op N;
  dirichlet_op D;
};

double f(const kwantxi::point& p){
  p.size();
  return 0;
}

double g(const kwantxi::point& p){
  if(p(1) == 0)
    return 1;
  if(p(1) == 1)
    return 2;
  return 0;  
}

using boost::shared_ptr;

int main(){
  
  gsl_set_error_handler(&kwantxi::errorHandler);

  try{
    using namespace std;
    using kwantxi::vector;
    shared_ptr<domain> 
      Omega(new domain("data/intr.v", "data/bdry.v", "data/nrml.v"));

    shared_ptr<overlapping_domain>
      O1(new overlapping_domain("data/intr1.v", 
				"data/bdry1.v", "data/nrml1.v")),
      O2(new overlapping_domain("data/intr2.v", 
				"data/bdry2.v", "data/nrml2.v"));
    set<shared_ptr<overlapping_domain> > dd;
    dd.insert(O1);
    dd.insert(O2);
    set_overlapper_info(dd);
     
    //Set some overlapper info manually for the corners.
    point p(2);


    
    //Define the rest of the bvp.
    
    shared_ptr<conv_diff> L(new conv_diff(100));
    
    //point n(2); n(1) = 1; n(2) = 0;
    //shared_ptr<smoothed_conv_diff> L(new smoothed_conv_diff(100, -0.001, n));
    
    shared_ptr<bdry_conv_diff> B(new bdry_conv_diff);
    
    shared_ptr<linear_BVP2> thebvp (new linear_BVP2(Omega,L,B,f,g));
    
    set<shared_ptr<const domain> > dd_in;
    dd_in.insert(O1);
    dd_in.insert(O2);
    
    piecewise_polynomial::set_n(7);
    thin_plate_spline::set_n(6);
    c_infty_kwantxi::set_epsilon(20.534);

    interpolator<piecewise_polynomial> solver(thebvp);
    
    //additive_schwarz_ddm<piecewise_polynomial> solver(dd_in, thebvp);
        
    size_t x_len = 20;
    size_t y_len = 12;
    matrix out(y_len, x_len);

    //Do the interior of the matrix.
    set<point>::iterator I = Omega->get_interior().begin();
    for(size_t j = 2; j <= x_len-1; j++){
      for(size_t i = y_len-1; i >= 2; i--){
	out(i,j) = solver.at(*I);
	I++;
      }
    }
    
    //Do the boundary of the matrix.
    I = Omega -> get_boundary().begin();
    for(size_t i = y_len; i >= 1; i--){
      out(i, 1) = solver.at(*I);
      I++;
    }
    for(size_t j = 2; j <= x_len-1; j++){
      out(y_len,j) = solver.at(*I); I++;
      out(1,j) = solver.at(*I); I++;
    }
    for(size_t i = y_len; i >= 1; i--){
      out(i,x_len) = solver.at(*I);
      I++;
    }
    

    matrix::set_precision(30);
    ofstream ofs("result.matrix");
    ofs << out;
    ofs.close();

    return 0;
    
  }
  catch(error exc){
    kwantxi::show_exception(exc);
  }
  return 0;
}

//Code for four subdomains (might want to put this elsewhere next time)

/*
  shared_ptr<overlapping_domain>
  O1(new overlapping_domain("data/intr1.v", 
  "data/bdry1.v", "data/nrml1.v")),
  O2(new overlapping_domain("data/intr2.v", 
  "data/bdry2.v", "data/nrml2.v")),
  O3(new overlapping_domain( "data/intr3.v", 
  "data/bdry3.v","data/nrml3.v")),
  O4(new overlapping_domain("data/intr4.v", 
  "data/bdry4.v", "data/nrml4.v"));
  set<shared_ptr<overlapping_domain> > dd;
  dd.insert(O1);
  dd.insert(O2);
  dd.insert(O3);
  dd.insert(O4);
  set_overlapper_info(dd);
    
  //Set some overlapper info manually for the corners.
  point p(2);
  p(1) = 0.384615384615385; p(2) = 0;
  O1 -> set_overlapper_info(p,O2);
  p(2) = 0.6;
  O1 -> set_overlapper_info(p,O2);

  p(1) = 0.205128205128205; p(2) = 0;
  O2 -> set_overlapper_info(p,O1);
  p(2) = 0.6;
  O2 -> set_overlapper_info(p,O1);

  p(1) = 0.58974358974359; p(2) = 0;
  O2 -> set_overlapper_info(p,O3);
  p(2) = 0.6;
  O2 -> set_overlapper_info(p,O3);

  p(1) = 0.41025641025641; p(2) = 0;
  O3 -> set_overlapper_info(p,O2);
  p(2) = 0.6;
  O3 -> set_overlapper_info(p,O2);

  p(1) = 0.794871794871795; p(2) = 0;
  O3 -> set_overlapper_info(p,O4);
  p(2) = 0.6;
  O3 -> set_overlapper_info(p,O4);

  p(1) = 0.615384615384615; p(2) = 0;
  O4 -> set_overlapper_info(p,O3);
  p(2) = 0.6;
  O4 -> set_overlapper_info(p,O3);
*/


//     p(1) = 0.923076923076923; p(2) = 0;
//     O1 -> set_overlapper_info(p,O2);
//     p(2) = 0.6;
//     O1 -> set_overlapper_info(p,O2);

//     p(1) = 0.794871794871795; p(2) = 0;
//     O2 -> set_overlapper_info(p,O1);
//     p(2) = 0.6;
//     O2 -> set_overlapper_info(p,O1);  

//For 20x12
/*
    p(1) = 0.368421052631579; p(2) = 0;
    O1 -> set_overlapper_info(p,O2);
    p(2) = 0.6;
    O1 -> set_overlapper_info(p,O2);
    
    p(1) = 0.210526315789474; p(2) = 0;
    O2 -> set_overlapper_info(p,O1);
    p(2) = 0.6;
    O2 -> set_overlapper_info(p,O1);
    
    p(1) = 0.578947368421053; p(2) = 0;
    O2 -> set_overlapper_info(p,O3);
    p(2) = 0.6;
    O2 -> set_overlapper_info(p,O3);
    
    p(1) = 0.421052631578947; p(2) = 0;
    O3 -> set_overlapper_info(p,O2);
    p(2) = 0.6;
    O3 -> set_overlapper_info(p,O2);
  
    p(1) = 0.789473684210526; p(2) = 0;
    O3 -> set_overlapper_info(p,O4);
    p(2) = 0.6;
    O3 -> set_overlapper_info(p,O4);
    
    p(1) = 0.631578947368421; p(2) = 0;
    O4 -> set_overlapper_info(p,O3);
    p(2) = 0.6;
    O4 -> set_overlapper_info(p,O3);
*/
